#!/usr/bin/env python

import sys
import pygame
import Box2D as box2d
import os

SCREEN_WIDTH = 1024
SCREEN_HEIGHT = 768


class block (object):
  def __init__ (self, world, size, pos, density = 0, friction = 0.0, sensor = False):
    body_def = box2d.b2BodyDef ()

    # position of body center
    body_def.position = (pos[0]+size[0]/2, pos[1]+size[1]/2)
    self.__body = world.CreateBody (body_def)

    shape_def = box2d.b2PolygonDef ()
    shape_def.SetAsBox (size[0]/2, size[1]/2)
    shape_def.isSensor = sensor
    self.sensor = sensor

    if 0 != density:
      shape_def.density = density
      shape_def.friction = friction

    self.__body.CreateShape (shape_def)

    if 0 != density:
      self.__body.SetMassFromShapes ()

    self.__body.SetUserData (self)

  def vertices (self):
    if 1 > len (self.__body.shapeList):
      return None

    v1 = self.__body.shapeList[0].vertices
    v2 = []
    for v in v1:
      v_ = self.__body.GetWorldPoint ((v[0], v[1]))
      v2.append ((v_.x, v_.y))
    return v2

  def force (self, vec):
    p = self.__body.GetWorldCenter ()
    self.__body.ApplyImpulse (vec, p)

  def draw (self, surf, color):
    v = self.vertices ()
    pygame.draw.polygon (surf, color, v, 2)

  def dupe (self, grid):
    body_def = box2d.b2BodyDef ()

    # position of body center
    body_def.position = (self.__body.position.x, self.__body.position.y)
    body_def.angle = self.__body.GetAngle ()
    body = grid.body (body_def)

    shape_def = box2d.b2PolygonDef ()
    shape_def.setVertices (self.__body.shapeList[0].asPolygon ().getVertices_tuple ())

    shape_def.density = self.__body.shapeList[0].GetDensity ()
    shape_def.friction = self.__body.shapeList[0].GetFriction ()
    shape_def.isSensor = self.sensor

    body.CreateShape (shape_def)

    if 0 != shape_def.density:
      body.SetMassFromShapes ()

    body.SetLinearVelocity (self.__body.GetLinearVelocity ())
    body.SetAngularVelocity (self.__body.GetAngularVelocity ())

    self.__body = body
    self.__body.SetUserData (self)


class goal (block):
  def __init__ (self, world, pos):
    super (goal, self).__init__ (world, (2*32, 3*32), pos, 0, 0, True)

  def draw (self, surf, color):
    v = self.vertices ()
    pygame.draw.polygon (surf, (0,255,0), v, 2)
    # door goes here


class ball:
  def __init__ (self, world, radius, pos, density = 0, friction = 0.0):
    body_def = box2d.b2BodyDef ()

    # position of body center
    body_def.position = (pos[0], pos[1])
    body_def.fixedRotation = True
    self.__body = world.body (body_def)

    shape_def = box2d.b2CircleDef ()
    shape_def.radius = radius
    self.sensor = False

    if 0 != density:
      shape_def.density = density
      shape_def.friction = friction

    self.__body.CreateShape (shape_def)

    if 0 != density:
      self.__body.SetMassFromShapes ()

    self.__body.SetUserData (self)

  def position (self):
    r = int (self.__body.shapeList[0].GetRadius ())
    return (int (self.__body.GetPosition ().x), int (self.__body.GetPosition ().y))

  def force (self, vec):
    p = self.__body.GetWorldCenter ()
    self.__body.ApplyImpulse (vec, p)

  def left (self, vec):
    p = self.__body.GetWorldCenter ()
    v = self.__body.GetLinearVelocity ()
    if v.x > -35:
      self.__body.ApplyImpulse (vec, p)

  def right (self, vec):
    p = self.__body.GetWorldCenter ()
    v = self.__body.GetLinearVelocity ()
    if v.x < 35:
      self.__body.ApplyImpulse (vec, p)

  def draw (self, surf, color):
    pass
    #r = int (self.__body.shapeList[0].GetRadius ())
    #pygame.draw.circle (surf, color, self.position (), r, 2)

  def dupe (self, grid):
    body_def = box2d.b2BodyDef ()

    # position of body center
    body_def.position = (self.__body.position.x, self.__body.position.y)
    body_def.angle = self.__body.GetAngle ()
    body_def.fixedRotation = True
    body = grid.body (body_def)

    shape_def = box2d.b2CircleDef ()
    shape_def.radius = self.__body.shapeList[0].GetRadius ()

    if 0 != self.__body.shapeList[0].GetDensity ():
      shape_def.density = self.__body.shapeList[0].GetDensity ()
      shape_def.friction = self.__body.shapeList[0].GetFriction ()

    body.CreateShape (shape_def)

    if 0 != self.__body.shapeList[0].GetDensity ():
      body.SetMassFromShapes ()

    body.SetLinearVelocity (self.__body.GetLinearVelocity ())
    body.SetAngularVelocity (self.__body.GetAngularVelocity ())

    self.__body = body
    body.SetUserData (self)



class grid:
  def __init__ (self, contact_listener, grid_size, color):
    self.__gsize = grid_size
    self.__color = color

    bounds = box2d.b2AABB ()
    bounds.lowerBound = (-50, -50)
    bounds.upperBound = ( SCREEN_WIDTH + 50,  SCREEN_HEIGHT + 50)
    self.__world = box2d.b2World (bounds, (0, 9.8), True)
    self.__contact_listener = contact_listener
    self.__world.SetContactListener (self.__contact_listener)

    # create ground
    block (self.__world, (SCREEN_WIDTH + 20, 20), (-10, SCREEN_HEIGHT + 10), 0.0, 0.9)
    block (self.__world, (SCREEN_WIDTH + 20, 20), (-10, SCREEN_HEIGHT-10), 90000, 0.9)
    # create walls
    block (self.__world, (20, SCREEN_HEIGHT - 20), (-10, 10), 0.0, 0.9)
    block (self.__world, (20, SCREEN_HEIGHT - 20), (SCREEN_WIDTH - 10, 10), 0.0, 0.9)
    #create ceiling
    block (self.__world, (SCREEN_WIDTH + 20, 20), (-10, -10), 0.0, 0.9)
    self.__goal = None

  def block (self, size, pos, weight = 90000):
    return block (self.__world,
                  (size[0]*self.__gsize, size[1]*self.__gsize),
                  (pos[0]*self.__gsize, pos[1]*self.__gsize),
                  weight,
                  0.9,
                 )
  def goal (self, pos):
    self.__goal = goal (self.__world, (pos[0]*self.__gsize, pos[1]*self.__gsize))

  def body (self, body_def):
    return self.__world.CreateBody (body_def)

  def step (self):
    timeStep = 3.0 / 60.0
    velocityIterations = 10
    positionIterations = 8

    self.__world.Step (timeStep, velocityIterations, positionIterations)

  
  def draw (self, surf):
    for body in self.__world.bodyList:
      if not body.GetUserData ():
        continue
      body.GetUserData ().draw (surf, self.__color)

  def pick (self, pos):
    aabb = box2d.b2AABB ()
    aabb.lowerBound.Set (pos[0]-0.05, pos[1]-0.05)
    aabb.upperBound.Set (pos[0]+0.05, pos[1]+0.05)
    return self.__world.Query (aabb, 1)

  def destroy (self, shape):
    self.__world.DestroyBody (shape.GetBody ())


class player:
  def __init__ (self, world, radius=20, pos=(3, 3), waittime = 60):
    self.__moving_left = False
    self.__moving_right = False
    self.__jumping = False    

    self.spriteball = ball (world, radius, (pos[0]*32, pos[1]*32), 1, 0.9)
    
    self.__picarray = []
    self.__picarray.append(pygame.image.load(os.path.join(os.path.dirname(sys.argv[0]),'data','images','cat1.png')))
    self.__picarray.append(pygame.image.load(os.path.join(os.path.dirname(sys.argv[0]),'data','images','cat2.png')))
    self.__standing = self.__picarray[0]
    
    self.__frame = 0
    self.__waittime = waittime
    self.__counter = 0
    
  def left (self, v):
    self.__moving_left = v
    # Make the counter equal to waittime so that the next key press immediately animates the character
    self.__counter = self.__waittime
        
  def right (self, v):
    self.__moving_right = v
    self.__counter = self.__waittime
    
  def jump (self, v):
    self.__jumping = v
 
  def move (self):
    if self.__moving_left:
      self.spriteball.left ((-999,0))
    if self.__moving_right:
      self.spriteball.right ((999,0))
    if self.__jumping:
      self.spriteball.force ((0,-24444))

  def __advanceframe (self):
    self.__frame = self.__frame + 1
    if self.__frame == len(self.__picarray):
      self.__frame = 0
          
  def __movement(self):
    if self.__counter == self.__waittime:
      self.__advanceframe()
      self.__counter = 0
    else:
      self.__counter = self.__counter + 1
      
  def animate(self):    
    if self.__moving_right == True:
      self.__movement()
    if self.__moving_left == True:
      self.__movement()

  def draw (self, surf):
    # Draws character without rotating it.  Either rotation needs to be stopped, or the character
    # has to be made to rotate with the block
    p = self.spriteball.position ()
    p = (p[0]-35, p[1]-33)
    # Draws character in both dimensions even though it only belongs in one.  Needs to be fixed
    if self.__moving_right == True:
      surf.blit(self.__picarray[self.__frame], p)
      self.__standing = self.__picarray[self.__frame]
    elif self.__moving_left == True:
      image = self.__picarray[self.__frame]
      image = pygame.transform.flip(image, True, False)
      surf.blit(image, p)
      self.__standing = image
    else:
      surf.blit(self.__standing, p)
    
    
class hud:

  def __init__ (self):
    self.__scarab = pygame.image.load(os.path.join(os.path.dirname(sys.argv[0]),'data','images','scarab32.gif'))
    self.__imagesize = 32
    self.__scarabpos = [SCREEN_WIDTH - 40, 20]

  def refresh (self, surf, dimleft, dimshiftallowed):
    if dimshiftallowed == True:
      for i in range(dimleft):      
        surf.blit(self.__scarab, self.__scarabpos)
        self.__scarabpos[0] = self.__scarabpos[0] - self.__imagesize      
      self.__scarabpos = [SCREEN_WIDTH - 40, 20] 
      
  


class game (box2d.b2ContactListener):
  def __init__ (self, level_file):
    super (game, self).__init__ ()

    self.done = False

    #self.__background = pygame.image.load(os.path.join(os.path.dirname(sys.argv[0]),'data','images','hall.bmp'))

    self.__dimleft = 9
    self.__dimshiftallowed = True

    self.__lite = grid (self, 32, (0,0,255))
    self.__dark = grid (self, 32, (255,0,0))
    self.__cworld = self.__lite
    self.__player = player (self.__cworld)
    self.__hud = hud ()    

    self.__jp = 0
    self.__level_done = False

    pygame.mixer.init ()
    pygame.mixer.music.load (os.path.join (os.path.dirname (sys.argv[0]),'data','music','Dark.ogg'))
    pygame.mixer.music.play (-1)

    self.__to_lite_sfx = pygame.mixer.Sound (os.path.join (os.path.dirname (sys.argv[0]),'data','sfx','Ahyay.wav'))
    self.__to_dark_sfx = pygame.mixer.Sound (os.path.join (os.path.dirname (sys.argv[0]),'data','sfx','Ohya.wav'))

    self.__dark_music_offset = 0
    self.__lite_music_offset = 0
    
    exec 'import ' + level_file + ' as level'
    level.init (self.__lite, self.__dark)    

  def switchworld (self):
    if self.__cworld == self.__lite:
      self.__cworld = self.__dark
      self.__to_dark_sfx.play ()
      self.__lite_music_offset += pygame.mixer.music.get_pos ()/1000
      pygame.mixer.music.load (os.path.join (os.path.dirname (sys.argv[0]),'data','music','Eerie.ogg'))
      pygame.mixer.music.play (-1, self.__dark_music_offset)
    else:
      self.__cworld = self.__lite
      self.__to_lite_sfx.play ()
      self.__dark_music_offset += pygame.mixer.music.get_pos ()/1000
      pygame.mixer.music.load (os.path.join (os.path.dirname (sys.argv[0]),'data','music','Dark.ogg'))
      pygame.mixer.music.play (-1, self.__lite_music_offset)

  def draw (self, surf):
    self.__player.move ()
    self.__lite.step ()
    self.__dark.step ()
    
    #surf.blit(self.__background, (0,0))
    self.__cworld.draw (surf)
    self.__player.draw (surf)
    self.__hud.refresh (surf, self.__dimleft, self.__dimshiftallowed)        
    self.__player.animate ()

  def handle_input (self, event):
    if event.type == pygame.KEYDOWN and event.key == pygame.K_LEFT:
      self.__player.left (True)
    if event.type == pygame.KEYDOWN and event.key == pygame.K_RIGHT:
      self.__player.right (True)
    if event.type == pygame.KEYDOWN and event.key == pygame.K_UP and self.__jp > 0:
      self.__player.jump (True)
    if event.type == pygame.KEYUP and event.key == pygame.K_LEFT:
      self.__player.left (False)
    if event.type == pygame.KEYUP and event.key == pygame.K_RIGHT:
      self.__player.right (False)
    if event.type == pygame.KEYUP and event.key == pygame.K_UP:
      self.__player.jump (False)

    if event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
      nine_times.warpblock (pygame.mouse.get_pos ())

    if event.type == pygame.MOUSEBUTTONDOWN and event.button == 3:
      nine_times.warpdimension ()

  def warpblock (self, pos):
    (count, shapes) = self.__cworld.pick (pos)
    if 1 == count:        
      actor = shapes[0].GetBody ().GetUserData ()      
      if self.__dimshiftallowed == True and \
         actor != self.__player.spriteball and \
         not actor.sensor:
        actor.dupe (self.oworld ())
        self.__cworld.destroy (shapes[0])
        # only shift blocks if there are scarabs left
        if self.__dimleft > 0:
          self.__dimleft = self.__dimleft - 1
        if self.__dimleft == 0:
          self.__dimshiftallowed = False

  def warpdimension (self) :
    (count, shapes) = self.__cworld.pick (self.__player.spriteball.position())
    if 1 == count:        
      actor = shapes[0].GetBody ().GetUserData ()
        # character block needs to be switched and destroyed no matter what
        #if self.__player.spriteball == actor:
      actor.dupe (self.oworld ())
      self.__cworld.destroy (shapes[0])      
      self.switchworld ()
        
  def oworld (self):
    if self.__lite == self.__cworld:
      return self.__dark
    else:
      return self.__lite

  def Add (self, point):
    if self.__player.spriteball == point.shape1.GetBody ().GetUserData () or \
       self.__player.spriteball == point.shape2.GetBody ().GetUserData ():
      #print "%.2f, %.2f" % (point.normal.x, point.normal.y)
      if point.normal.y < -0.9:
        self.__jp += 1

      if point.shape1.GetBody ().GetUserData ().sensor or \
         point.shape2.GetBody ().GetUserData ().sensor:
        self.__level_done = True

  def Persist (self, point):
    pass

  def Remove (self, point):
    if self.__player.spriteball == point.shape1.GetBody ().GetUserData () or \
       self.__player.spriteball == point.shape2.GetBody ().GetUserData ():
      #print "%.2f, %.2f" % (point.normal.x, point.normal.y)
      if point.normal.y < -0.9:
        self.__jp -= 1

      if 1 > self.__jp:
        self.__jp = 0
        self.__player.jump (False)

  def Result (self, point):
    pass

  def is_level_done (self):
    return self.__level_done

  def gameover (self):
    self.done = True
    

    

if __name__ == '__main__':
  level = 1
  if len(sys.argv) > 1:
    level = int (sys.argv[1])
  
  pygame.init ()
  flags = pygame.DOUBLEBUF | pygame.HWSURFACE | pygame.FULLSCREEN
  screen = pygame.display.set_mode ((SCREEN_WIDTH, SCREEN_HEIGHT), flags)
  pygame.display.set_caption ('Nine Times')

  prefix = 'data.levels.level0'
  nine_times = game (prefix + str (level))

  while not nine_times.done:

    screen.fill ((95,79,55))
    nine_times.draw (screen)
    pygame.display.flip ()

    event = pygame.event.poll ()

    if event.type == pygame.QUIT:
      nine_times.done = True

    if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
      nine_times.done = True

    nine_times.handle_input (event)

    if nine_times.is_level_done ():
      level += 1
      nine_times = game (prefix + str (level))

